/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996 by Kevin P. Lawton.

BOCHS is shareware for PERSONAL USE only.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/



#define BX_ARITH_C 1

#include "bochs.h"


  INLINE void
bx_ADD_EbGb()
{
  Bit32u op1_addr;
  unsigned op2_addr, op1_type;
  bx_segment_reg_t *op1_seg_reg;

  Bit8u op2, op1, sum;
  Bit8u op2_b7, op1_b7, sum_b7;

  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op2 is a register, op2_addr is an index of a register */
  op2 = BX_READ_8BIT_REG(op2_addr);

  /* op1 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1);
    BX_HANDLE_EXCEPTION()
    }

  sum = op1 + op2;

  /* now write sum back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, sum);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &sum);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * ADD affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of =  (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);
}


  INLINE void
bx_ADD_EvGv()
{
  Bit32u op1_addr;
  unsigned op1_type, op2_addr;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, sum_b31;
    Bit32u op2_32, op1_32, sum_32;

    /* op2_32 is a register, op2_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(op2_addr);

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, sum_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &sum_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;
    Bit16u op2_b15, op1_b15, sum_b15;


    /* op2_16 is a register, op2_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(op2_addr);

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    sum_16 = op1_16 + op2_16;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, sum_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &sum_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}


  INLINE void
bx_ADD_GbEb()
{
  Bit8u op1, op2, sum;
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;
  Bit8u op1_b7, op2_b7, sum_b7;



  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op1 is a register, op1_addr is an index of a register */
  op1 = BX_READ_8BIT_REG(op1_addr);

  /* op2 is a register or memory reference */
  if (op2_type == BX_REGISTER_REF) {
    op2 = BX_READ_8BIT_REG(op2_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op2_seg_reg, op2_addr, &op2);
    BX_HANDLE_EXCEPTION()
    }

  sum = op1 + op2;

  /* now write sum back to destination, which is a register */
  BX_WRITE_8BIT_REG(op1_addr, sum);

  /* set eflags:
   * ADD affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of = (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);
}

  INLINE void
bx_ADD_GvEv()
{
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, sum_b31;
    Bit32u op1_32, op2_32, sum_32;

    /* op1_32 is a register, op1_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(op1_addr);

    /* op2_32 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op2_seg_reg, op2_addr, &op2_32);
      BX_HANDLE_EXCEPTION()
      }

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    BX_WRITE_32BIT_REG(op1_addr, sum_32);

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, sum_16;
    Bit16u op1_b15, op2_b15, sum_b15;


    /* op1_16 is a register, op1_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(op1_addr);

    /* op2_16 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op2_seg_reg, op2_addr, &op2_16);
      BX_HANDLE_EXCEPTION()
      }

    sum_16 = op1_16 + op2_16;
    /* now write sum back to destination */

    BX_WRITE_16BIT_REG(op1_addr, sum_16);

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}

  INLINE void
bx_ADD_ALIb()
{
  Bit8u op1, op2, sum;
  Bit8u op1_b7, op2_b7, sum_b7;


  op1 = AL;

  op2 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  sum = op1 + op2;

  /* now write sum back to destination, which is a register */
  AL = sum;

  /* set eflags:
   * ADD affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of = (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);
}

  INLINE void
bx_ADD_eAXIv()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, sum_32;
    Bit32u op1_b31, op2_b31, sum_b31;

    op1_32 = EAX;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    EAX = sum_32;

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, sum_16;
    Bit16u op1_b15, op2_b15, sum_b15;

    op1_16 = AX;

    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    sum_16 = op1_16 + op2_16;

    /* now write sum back to destination */
    AX = sum_16;

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}

  INLINE void
bx_ADC_EbGb()
{
  Bit8u op2, op1, sum;
  Bit32u op1_addr;
  unsigned op1_type, op2_addr;
  bx_segment_reg_t *op1_seg_reg;
  Bit8u op2_b7, op1_b7, sum_b7;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op2 is a register, op2_addr is an index of a register */
  op2 = BX_READ_8BIT_REG(op2_addr);

  /* op1 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1);
    BX_HANDLE_EXCEPTION()
    }

  sum = op1 + op2 + CF;


  /* now write sum back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, sum);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &sum);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * ADC affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of =  (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1) || (CF && sum==op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);
}

  INLINE void
bx_ADC_EvGv()
{
  Bit32u op1_addr;
  unsigned op1_type, op2_addr;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, sum_b31;
    Bit32u op2_32, op1_32, sum_32;

    /* op2_32 is a register, op2_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(op2_addr);

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    sum_32 = op1_32 + op2_32 + CF;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, sum_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &sum_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32) || (CF && sum_32==op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;
    Bit16u op2_b15, op1_b15, sum_b15;


    /* op2_16 is a register, op2_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(op2_addr);

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    sum_16 = op1_16 + op2_16 + CF;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, sum_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &sum_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16) || (CF && sum_16==op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}

  INLINE void
bx_ADC_GbEb()
{
  Bit8u op1, op2, sum;
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;
  Bit8u op1_b7, op2_b7, sum_b7;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op1 is a register, op1_addr is an index of a register */
  op1 = BX_READ_8BIT_REG(op1_addr);

  /* op2 is a register or memory reference */
  if (op2_type == BX_REGISTER_REF) {
    op2 = BX_READ_8BIT_REG(op2_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op2_seg_reg, op2_addr, &op2);
    BX_HANDLE_EXCEPTION()
    }

  sum = op1 + op2 + CF;

  /* set eflags:
   * ADC affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of = (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1) || (CF && sum==op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);

  /* now write sum back to destination, which is a register */
  BX_WRITE_8BIT_REG(op1_addr, sum);
}

  INLINE void
bx_ADC_GvEv()
{
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, sum_b31;
    Bit32u op1_32, op2_32, sum_32;

    /* op1_32 is a register, op1_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(op1_addr);

    /* op2_32 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op2_seg_reg, op2_addr, &op2_32);
      BX_HANDLE_EXCEPTION()
      }

    sum_32 = op1_32 + op2_32 + CF;

    /* now write sum back to destination */
    BX_WRITE_32BIT_REG(op1_addr, sum_32);

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32) || (CF && sum_32==op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, sum_16;
    Bit16u op1_b15, op2_b15, sum_b15;


    /* op1_16 is a register, op1_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(op1_addr);

    /* op2_16 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op2_seg_reg, op2_addr, &op2_16);
      BX_HANDLE_EXCEPTION()
      }

    sum_16 = op1_16 + op2_16 + CF;

    /* now write sum back to destination */
    BX_WRITE_16BIT_REG(op1_addr, sum_16);

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16) || (CF && sum_16==op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}

  INLINE void
bx_ADC_ALIb()
{
  Bit8u op1, op2, sum;
  Bit8u op1_b7, op2_b7, sum_b7;


  op1 = AL;

  op2 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  sum = op1 + op2 + CF;

  /* now write sum back to destination, which is a register */
  AL = sum;

  /* set eflags:
   * ADC affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of = (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1) || (CF && sum==op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);
}

  INLINE void
bx_ADC_eAXIv()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, sum_b31;
    Bit32u op1_32, op2_32, sum_32;

    op1_32 = EAX;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    sum_32 = op1_32 + op2_32 + CF;

    /* now write sum back to destination */
    EAX = sum_32;

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32) || (CF && sum_32==op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, sum_16;
    Bit16u op1_b15, op2_b15, sum_b15;

    op1_16 = AX;

    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    sum_16 = op1_16 + op2_16 + CF;

    /* now write sum back to destination */
    AX = sum_16;

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16) || (CF && sum_16==op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}

  INLINE void
bx_SBB_EbGb()
{
  Bit8u op2_8, op1_8, diff_8;
  Bit32u op1_addr;
  unsigned op1_type, op2_addr;
  bx_segment_reg_t *op1_seg_reg;
  Bit8u op2_b7, op1_b7, diff_b7;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op2 is a register, op2_addr is an index of a register */
  op2_8 = BX_READ_8BIT_REG(op2_addr);

  /* op1_8 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - (op2_8 + CF);

  /* now write diff back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, diff_8);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &diff_8);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * SBB affects the following flags: O,S,Z,A,P,C
   */
  op1_b7  = op1_8  & 0x80;
  op2_b7  = op2_8  & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of = (op1_b7 ^ op2_b7) && (op1_b7 ^ diff_b7);
  bx_cpu.eflags.cf = (op1_8 < diff_8) || ((op2_8==0xff) && CF);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_SBB_EvGv()
{
  Bit32u op1_addr;
  unsigned op1_type, op2_addr;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    /* op2_32 is a register, op2_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(op2_addr);

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - (op2_32 + CF);

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, diff_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &diff_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < diff_32) || ((op2_32==0xffffffff) && CF);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    /* op2_16 is a register, op2_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(op2_addr);

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - (op2_16 + CF);

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, diff_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &diff_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < diff_b15) || ((op2_16==0xffff) && CF);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_SBB_GbEb()
{
  Bit8u op1_8, op2_8, diff_8;
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;
  Bit8u op1_b7, op2_b7, diff_b7;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op1 is a register, op1_addr is an index of a register */
  op1_8 = BX_READ_8BIT_REG(op1_addr);

  /* op2 is a register or memory reference */
  if (op2_type == BX_REGISTER_REF) {
    op2_8 = BX_READ_8BIT_REG(op2_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op2_seg_reg, op2_addr, &op2_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - (op2_8 + CF);

  /* now write diff back to destination, which is a register */
  BX_WRITE_8BIT_REG(op1_addr, diff_8);

  /* set eflags:
   * SBB affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1_8 & 0x80;
  op2_b7 = op2_8 & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of = (op1_b7 ^ op2_b7) && (op1_b7 ^ diff_b7);
  bx_cpu.eflags.cf = (op1_8 < diff_8) || ((op2_8==0xff) && CF);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_SBB_GvEv()
{
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, diff_b31;
    Bit32u op1_32, op2_32, diff_32;

    /* op1_32 is a register, op1_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(op1_addr);

    /* op2_32 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op2_seg_reg, op2_addr, &op2_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - (op2_32 + CF);

    /* now write diff back to destination */
    BX_WRITE_32BIT_REG(op1_addr, diff_32);

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < diff_32) || ((op2_32==0xffffffff) && CF);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;
    Bit16u op1_b15, op2_b15, diff_b15;


    /* op1_16 is a register, op1_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(op1_addr);

    /* op2_16 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op2_seg_reg, op2_addr, &op2_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - (op2_16 + CF);

    /* now write diff back to destination */
    BX_WRITE_16BIT_REG(op1_addr, diff_16);

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < diff_16) || ((op2_16==0xffff) && CF);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_SBB_ALIb()
{
  Bit8u op1_8, op2_8, diff_8;
  Bit8u op1_b7, op2_b7, diff_b7;


  op1_8 = AL;

  op2_8 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  diff_8 = op1_8 - (op2_8 + CF);

  /* now write diff back to destination, which is a register */
  AL = diff_8;

  /* set eflags:
   * SBB affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1_8 & 0x80;
  op2_b7 = op2_8 & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of = (op1_b7 ^ op2_b7) && (op1_b7 ^ diff_b7);
  bx_cpu.eflags.cf = (op1_8 < diff_8) || ((op2_8==0xff) && CF);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_SBB_eAXIv()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, diff_b31;
    Bit32u op1_32, op2_32, diff_32;

    op1_32 = EAX;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    diff_32 = op1_32 - (op2_32 + CF);

    /* now write diff back to destination */
    EAX = diff_32;

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < diff_32) || ((op2_32==0xffffffff) && CF);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;
    Bit16u op1_b15, op2_b15, diff_b15;

    op1_16 = AX;

    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    diff_16 = op1_16 - (op2_16 + CF);

    /* now write diff back to destination */
    AX = diff_16;

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < diff_16) || ((op2_16==0xffff) && CF);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}


  INLINE void
bx_SBB_EbIb()
{
  Bit8u op2_8, op1_8, diff_8;
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;
  Bit8u op2_b7, op1_b7, diff_b7;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  op2_8 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  /* op1_8 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - (op2_8 + CF);

  /* now write diff back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, diff_8);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &diff_8);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * SBB affects the following flags: O,S,Z,A,P,C
   */
  op1_b7  = op1_8  & 0x80;
  op2_b7  = op2_8  & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of =  (op1_b7 ^ op2_b7) && (op1_b7 ^ diff_b7);
  bx_cpu.eflags.cf = (op1_8 < diff_8) || ((op2_8==0xff) && CF);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}


  INLINE void
bx_SBB_EvIv()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - (op2_32 + CF);

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, diff_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &diff_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < diff_32) || ((op2_32==0xffffffff) && CF);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - (op2_16 + CF);

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, diff_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &diff_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < diff_16) || ((op2_16==0xffff) && CF);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_SBB_EvIb()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - (op2_32 + CF);

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, diff_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &diff_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < diff_32) || ((op2_32==0xffffffff) && CF);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    op2_16 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - (op2_16 + CF);

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, diff_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &diff_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SBB affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < diff_16) || ((op2_16==0xffff) && CF);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}



  INLINE void
bx_SUB_EbGb()
{
  Bit8u op2_8, op1_8, diff_8;
  Bit32u op1_addr;
  unsigned op2_addr, op1_type;
  bx_segment_reg_t *op1_seg_reg;
  Bit8u op2_b7, op1_b7, diff_b7;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op2 is a register, op2_addr is an index of a register */
  op2_8 = BX_READ_8BIT_REG(op2_addr);

  /* op1_8 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - op2_8;

  /* now write diff back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, diff_8);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &diff_8);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * SUB affects the following flags: O,S,Z,A,P,C
   */
  op1_b7  = op1_8  & 0x80;
  op2_b7  = op2_8  & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of =  (op1_b7 ^ op2_b7) && (op1_b7 ^ diff_b7);
  bx_cpu.eflags.cf = (op1_8 < op2_8);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_SUB_EvGv()
{
  Bit32u op1_addr;
  unsigned op1_type, op2_addr;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    /* op2_32 is a register, op2_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(op2_addr);

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, diff_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &diff_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    /* op2_16 is a register, op2_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(op2_addr);

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - op2_16;

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, diff_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &diff_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;


    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_SUB_GbEb()
{
  Bit8u op1_8, op2_8, diff_8;
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;
  Bit8u op1_b7, op2_b7, diff_b7;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op1 is a register, op1_addr is an index of a register */
  op1_8 = BX_READ_8BIT_REG(op1_addr);

  /* op2 is a register or memory reference */
  if (op2_type == BX_REGISTER_REF) {
    op2_8 = BX_READ_8BIT_REG(op2_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op2_seg_reg, op2_addr, &op2_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - op2_8;

  /* now write diff back to destination, which is a register */
  BX_WRITE_8BIT_REG(op1_addr, diff_8);

  /* set eflags:
   * SUB affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1_8 & 0x80;
  op2_b7 = op2_8 & 0x80;
  diff_b7 = diff_8 & 0x80;


  bx_cpu.eflags.of = (op1_b7 ^ op2_b7) && (op1_b7 ^ diff_b7);
  bx_cpu.eflags.cf = (op1_8 < op2_8);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_SUB_GvEv()
{
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, diff_b31;
    Bit32u op1_32, op2_32, diff_32;

    /* op1_32 is a register, op1_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(op1_addr);

    /* op2_32 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op2_seg_reg, op2_addr, &op2_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    BX_WRITE_32BIT_REG(op1_addr, diff_32);

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;
    Bit16u op1_b15, op2_b15, diff_b15;


    /* op1_16 is a register, op1_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(op1_addr);

    /* op2_16 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op2_seg_reg, op2_addr, &op2_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - op2_16;

    /* now write diff back to destination */
    BX_WRITE_16BIT_REG(op1_addr, diff_16);

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_SUB_ALIb()
{
  Bit8u op1_8, op2_8, diff_8;
  Bit8u op1_b7, op2_b7, diff_b7;

  op1_8 = AL;

  op2_8 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  diff_8 = op1_8 - op2_8;

  /* now write diff back to destination, which is a register */
  AL = diff_8;

  /* set eflags:
   * SUB affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1_8 & 0x80;
  op2_b7 = op2_8 & 0x80;
  diff_b7 = diff_8 & 0x80;


  bx_cpu.eflags.of = (op1_b7 ^ op2_b7) && (op1_b7 ^ diff_b7);
  bx_cpu.eflags.cf = (op1_8 < op2_8);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_SUB_eAXIv()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, diff_b31;
    Bit32u op1_32, op2_32, diff_32;

    op1_32 = EAX;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    EAX = diff_32;

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;
    Bit16u op1_b15, op2_b15, diff_b15;

    op1_16 = AX;

    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    diff_16 = op1_16 - op2_16;


    /* now write diff back to destination */
    AX = diff_16;

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}


  INLINE void
bx_CMP_EbGb()
{
  Bit8u op2_8, op1_8, diff_8;
  Bit32u op1_addr;
  unsigned op1_type, op2_addr;
  bx_segment_reg_t *op1_seg_reg;
  Bit8u op2_b7, op1_b7, diff_b7;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op2 is a register, op2_addr is an index of a register */
  op2_8 = BX_READ_8BIT_REG(op2_addr);

  /* op1_8 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - op2_8;

  /* set eflags:
   * CMP affects the following flags: O,S,Z,A,P,C
   */
  op1_b7  = op1_8  & 0x80;
  op2_b7  = op2_8  & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of =  (op1_b7 ^ op2_b7) && (diff_b7 ^ op1_b7);
  bx_cpu.eflags.cf = (op1_8 < op2_8);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_CMP_EvGv()
{
  Bit32u op1_addr;
  unsigned op1_type, op2_addr;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    /* op2_32 is a register, op2_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(op2_addr);

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - op2_32;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (diff_b31 ^ op1_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    /* op2_16 is a register, op2_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(op2_addr);

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - op2_16;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (diff_b15 ^ op1_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_CMP_GbEb()
{
  Bit8u op1_8, op2_8, diff_8;
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;
  Bit8u op1_b7, op2_b7, diff_b7;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op1 is a register, op1_addr is an index of a register */
  op1_8 = BX_READ_8BIT_REG(op1_addr);

  /* op2 is a register or memory reference */
  if (op2_type == BX_REGISTER_REF) {
    op2_8 = BX_READ_8BIT_REG(op2_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op2_seg_reg, op2_addr, &op2_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - op2_8;

  /* set eflags:
   * CMP affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1_8 & 0x80;
  op2_b7 = op2_8 & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of = (op1_b7 ^ op2_b7) && (diff_b7 ^ op1_b7);
  bx_cpu.eflags.cf = (op1_8 < op2_8);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_CMP_GvEv()
{
  Bit32u op2_addr;
  unsigned op1_addr, op2_type;
  bx_segment_reg_t *op2_seg_reg;


  bx_decode_exgx(&op1_addr, &op2_addr, &op2_type, &op2_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, diff_b31;
    Bit32u op1_32, op2_32, diff_32;

    /* op1_32 is a register, op1_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(op1_addr);

    /* op2_32 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op2_seg_reg, op2_addr, &op2_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - op2_32;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (diff_b31 ^ op1_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;
    Bit16u op1_b15, op2_b15, diff_b15;


    /* op1_16 is a register, op1_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(op1_addr);

    /* op2_16 is a register or memory reference */
    if (op2_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(op2_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op2_seg_reg, op2_addr, &op2_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - op2_16;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (diff_b15 ^ op1_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}


  INLINE void
bx_CMP_ALIb()
{
  Bit8u op1_8, op2_8, diff_8;
  Bit8u op1_b7, op2_b7, diff_b7;


  op1_8 = AL;

  op2_8 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  diff_8 = op1_8 - op2_8;

  /* set eflags:
   * CMP affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1_8 & 0x80;
  op2_b7 = op2_8 & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of = (op1_b7 ^ op2_b7) && (diff_b7 ^ op1_b7);
  bx_cpu.eflags.cf = (op1_8 < op2_8);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_CMP_eAXIv()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_b31, op2_b31, diff_b31;
    Bit32u op1_32, op2_32, diff_32;

    op1_32 = EAX;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    diff_32 = op1_32 - op2_32;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (diff_b31 ^ op1_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;
    Bit16u op1_b15, op2_b15, diff_b15;

    op1_16 = AX;

    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    diff_16 = op1_16 - op2_16;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (diff_b15 ^ op1_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_INC_eAX()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u eax;

    eax = EAX;

    eax++;

    EAX = eax;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (eax == 0x80000000);
    bx_cpu.eflags.sf = (eax >= 0x80000000);
    bx_cpu.eflags.zf = (eax == 0);
    bx_cpu.eflags.af = (eax & 0x0f) == 0;
    SET_PF(eax);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u ax;

    ax = AX;

    ax++;

    AX = ax;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (ax == 0x8000);
    bx_cpu.eflags.sf = (ax >= 0x8000);
    bx_cpu.eflags.zf = (ax == 0);
    bx_cpu.eflags.af = (ax & 0x0f) == 0;
    SET_PF(ax);
    }
}

  INLINE void
bx_INC_eCX()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ecx;

    ecx = ECX;

    ecx++;

    ECX = ecx;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (ecx == 0x80000000);
    bx_cpu.eflags.sf = (ecx >= 0x80000000);
    bx_cpu.eflags.zf = (ecx == 0);
    bx_cpu.eflags.af = (ecx & 0x0f) == 0;
    SET_PF(ecx);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u cx;

    cx = CX;

    cx++;

    CX = cx;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (cx == 0x8000);
    bx_cpu.eflags.sf = (cx >= 0x8000);
    bx_cpu.eflags.zf = (cx == 0);
    bx_cpu.eflags.af = (cx & 0x0f) == 0;
    SET_PF(cx);
    }
}

  INLINE void
bx_INC_eDX()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u edx;

    edx = EDX;

    edx++;

    EDX = edx;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (edx == 0x80000000);
    bx_cpu.eflags.sf = (edx >= 0x80000000);
    bx_cpu.eflags.zf = (edx == 0);
    bx_cpu.eflags.af = (edx & 0x0f) == 0;
    SET_PF(edx);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u dx;

    dx = DX;

    dx++;

    DX = dx;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (dx == 0x8000);
    bx_cpu.eflags.sf = (dx >= 0x8000);
    bx_cpu.eflags.zf = (dx == 0);
    bx_cpu.eflags.af = (dx & 0x0f) == 0;
    SET_PF(dx);
    }
}

  INLINE void
bx_INC_eBX()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ebx;

    ebx = EBX;

    ebx++;

    EBX = ebx;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (ebx == 0x80000000);
    bx_cpu.eflags.sf = (ebx >= 0x80000000);
    bx_cpu.eflags.zf = (ebx == 0);
    bx_cpu.eflags.af = (ebx & 0x0f) == 0;
    SET_PF(ebx);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u bx;

    bx = BX;

    bx++;

    BX = bx;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (bx == 0x8000);
    bx_cpu.eflags.sf = (bx >= 0x8000);
    bx_cpu.eflags.zf = (bx == 0);
    bx_cpu.eflags.af = (bx & 0x0f) == 0;
    SET_PF(bx);
    }
}

  INLINE void
bx_INC_eSP()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u esp;

    esp = ESP;

    esp++;

    ESP = esp;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (esp == 0x80000000);
    bx_cpu.eflags.sf = (esp >= 0x80000000);
    bx_cpu.eflags.zf = (esp == 0);
    bx_cpu.eflags.af = (esp & 0x0f) == 0;
    SET_PF(esp);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u sp;

    sp = SP;

    sp++;

    SP = sp;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (sp == 0x8000);
    bx_cpu.eflags.sf = (sp >= 0x8000);
    bx_cpu.eflags.zf = (sp == 0);
    bx_cpu.eflags.af = (sp & 0x0f) == 0;
    SET_PF(sp);
    }
}

  INLINE void
bx_INC_eBP()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ebp;

    ebp = EBP;

    ebp++;

    EBP = ebp;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (ebp == 0x80000000);
    bx_cpu.eflags.sf = (ebp >= 0x80000000);
    bx_cpu.eflags.zf = (ebp == 0);
    bx_cpu.eflags.af = (ebp & 0x0f) == 0;
    SET_PF(ebp);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u bp;

    bp = BP;

    bp++;

    BP = bp;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (bp == 0x8000);
    bx_cpu.eflags.sf = (bp >= 0x8000);
    bx_cpu.eflags.zf = (bp == 0);
    bx_cpu.eflags.af = (bp & 0x0f) == 0;
    SET_PF(bp);
    }
}

  INLINE void
bx_INC_eSI()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u esi;

    esi = ESI;

    esi++;

    ESI = esi;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (esi == 0x80000000);
    bx_cpu.eflags.sf = (esi >= 0x80000000);
    bx_cpu.eflags.zf = (esi == 0);
    bx_cpu.eflags.af = (esi & 0x0f) == 0;
    SET_PF(esi);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u si;

    si = SI;

    si++;

    SI = si;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (si == 0x8000);
    bx_cpu.eflags.sf = (si >= 0x8000);
    bx_cpu.eflags.zf = (si == 0);
    bx_cpu.eflags.af = (si & 0x0f) == 0;
    SET_PF(si);
    }
}

  INLINE void
bx_INC_eDI()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u edi;

    edi = EDI;

    edi++;

    EDI = edi;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (edi == 0x80000000);
    bx_cpu.eflags.sf = (edi >= 0x80000000);
    bx_cpu.eflags.zf = (edi == 0);
    bx_cpu.eflags.af = (edi & 0x0f) == 0;
    SET_PF(edi);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u di;

    di = DI;

    di++;

    DI = di;

    /* INC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (di == 0x8000);
    bx_cpu.eflags.sf = (di >= 0x8000);
    bx_cpu.eflags.zf = (di == 0);
    bx_cpu.eflags.af = (di & 0x0f) == 0;
    SET_PF(di);
    }
}

  INLINE void
bx_DEC_eAX()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u eax;

    eax = EAX;

    eax--;

    EAX = eax;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (eax == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (eax >= 0x80000000);
    bx_cpu.eflags.zf = (eax == 0);
    bx_cpu.eflags.af = (eax & 0x0f) == 0x0f;
    SET_PF(eax);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u ax;

    ax = AX;

    ax--;

    AX = ax;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (ax == 0x7FFF);
    bx_cpu.eflags.sf = (ax >= 0x8000);
    bx_cpu.eflags.zf = (ax == 0);
    bx_cpu.eflags.af = (ax & 0x0f) == 0x0f;
    SET_PF(ax);
    }
}

  INLINE void
bx_DEC_eCX()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ecx;

    ecx = ECX;

    ecx--;

    ECX = ecx;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (ecx == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (ecx >= 0x80000000);
    bx_cpu.eflags.zf = (ecx == 0);
    bx_cpu.eflags.af = (ecx & 0x0f) == 0x0f;
    SET_PF(ecx);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u cx;

    cx = CX;

    cx--;

    CX = cx;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (cx == 0x7FFF);
    bx_cpu.eflags.sf = (cx >= 0x8000);
    bx_cpu.eflags.zf = (cx == 0);
    bx_cpu.eflags.af = (cx & 0x0f) == 0x0f;
    SET_PF(cx);
    }
}

  INLINE void
bx_DEC_eDX()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u edx;

    edx = EDX;

    edx--;

    EDX = edx;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (edx == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (edx >= 0x80000000);
    bx_cpu.eflags.zf = (edx == 0);
    bx_cpu.eflags.af = (edx & 0x0f) == 0x0f;
    SET_PF(edx);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u dx;

    dx = DX;

    dx--;

    DX = dx;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (dx == 0x7FFF);
    bx_cpu.eflags.sf = (dx >= 0x8000);
    bx_cpu.eflags.zf = (dx == 0);
    bx_cpu.eflags.af = (dx & 0x0f) == 0x0f;
    SET_PF(dx);
    }
}

  INLINE void
bx_DEC_eBX()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ebx;

    ebx = EBX;

    ebx--;

    EBX = ebx;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (ebx == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (ebx >= 0x80000000);
    bx_cpu.eflags.zf = (ebx == 0);
    bx_cpu.eflags.af = (ebx & 0x0f) == 0x0f;
    SET_PF(ebx);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u bx;

    bx = BX;

    bx--;

    BX = bx;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (bx == 0x7FFF);
    bx_cpu.eflags.sf = (bx >= 0x8000);
    bx_cpu.eflags.zf = (bx == 0);
    bx_cpu.eflags.af = (bx & 0x0f) == 0x0f;
    SET_PF(bx);
    }
}

  INLINE void
bx_DEC_eSP()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u esp;

    esp = ESP;

    esp--;

    ESP = esp;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (esp == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (esp >= 0x80000000);
    bx_cpu.eflags.zf = (esp == 0);
    bx_cpu.eflags.af = (esp & 0x0f) == 0x0f;
    SET_PF(esp);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u sp;

    sp = SP;

    sp--;

    SP = sp;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (sp == 0x7FFF);
    bx_cpu.eflags.sf = (sp >= 0x8000);
    bx_cpu.eflags.zf = (sp == 0);
    bx_cpu.eflags.af = (sp & 0x0f) == 0x0f;
    SET_PF(sp);
    }
}

  INLINE void
bx_DEC_eBP()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ebp;

    ebp = EBP;

    ebp--;

    EBP = ebp;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (ebp == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (ebp >= 0x80000000);
    bx_cpu.eflags.zf = (ebp == 0);
    bx_cpu.eflags.af = (ebp & 0x0f) == 0x0f;
    SET_PF(ebp);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u bp;

    bp = BP;

    bp--;

    BP = bp;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (bp == 0x7FFF);
    bx_cpu.eflags.sf = (bp >= 0x8000);
    bx_cpu.eflags.zf = (bp == 0);
    bx_cpu.eflags.af = (bp & 0x0f) == 0x0f;
    SET_PF(bp);
    }
}

  INLINE void
bx_DEC_eSI()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u esi;

    esi = ESI;

    esi--;

    ESI = esi;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (esi == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (esi >= 0x80000000);
    bx_cpu.eflags.zf = (esi == 0);
    bx_cpu.eflags.af = (esi & 0x0f) == 0x0f;
    SET_PF(esi);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u si;

    si = SI;

    si--;

    SI = si;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (si == 0x7FFF);
    bx_cpu.eflags.sf = (si >= 0x8000);
    bx_cpu.eflags.zf = (si == 0);
    bx_cpu.eflags.af = (si & 0x0f) == 0x0f;
    SET_PF(si);
    }
}

  INLINE void
bx_DEC_eDI()
{
#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u edi;

    edi = EDI;

    edi--;

    EDI = edi;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (edi == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (edi >= 0x80000000);
    bx_cpu.eflags.zf = (edi == 0);
    bx_cpu.eflags.af = (edi & 0x0f) == 0x0f;
    SET_PF(edi);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u di;

    di = DI;

    di--;

    DI = di;

    /* DEC effects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (di == 0x7FFF);
    bx_cpu.eflags.sf = (di >= 0x8000);
    bx_cpu.eflags.zf = (di == 0);
    bx_cpu.eflags.af = (di & 0x0f) == 0x0f;
    SET_PF(di);
    }
}


  INLINE void
bx_CBW()
{
  /* CBW: no flags are effected */

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* CWDE opcode */
    EAX = (Bit16s) AX;
    }
  else
#endif /* BX_CPU > 2 */
    { /* CBW opcode */
    AX = (Bit8s) AL;
    }
}

  INLINE void
bx_CWD()
{
  /* CWD: no flags are affected */

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* CDQ opcode */
    if (EAX & 0x80000000) {
      EDX = 0xFFFFFFFF;
      }
    else {
      EDX = 0x00000000;
      }
    }
  else
#endif /* BX_CPU > 2 */
    { /* CWD opcode */
    if (AX & 0x8000) {
      DX = 0xFFFF;
      }
    else {
      DX = 0x0000;
      }
    }
}

  INLINE void
bx_CMPXCHG_XBTS()
{
#if BX_CPU < 4
  bx_printf("CMPXCHG_XBTS:\n");
  bx_exception(6, 0, 0);
  return;
#else
  bx_panic("CMPXCHG_XBTS:\n");
#endif
}

  INLINE void
bx_CMPXCHG_IBTS()
{
  bx_panic("CMPXCHG_IBTS:\n");
}

  INLINE void
bx_XADD_EbGb()
{
#if BX_CPU < 4
  bx_panic("XADD_EbGb: not supported on < 80486\n");
#else

  Bit8u op2, op1, sum;
  Bit32u op1_addr;
  unsigned op2_addr, op1_type;
  bx_segment_reg_t *op1_seg_reg;
  Bit8u op2_b7, op1_b7, sum_b7;

  /* XADD dst(r/m8), src(r8)
   * temp <-- src + dst         | sum = op2 + op1
   * src  <-- dst               | op2 = op1
   * dst  <-- tmp               | op1 = sum
   */

  bx_decode_exgx(&op2_addr, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op2 is a register, op2_addr is an index of a register */
  op2 = BX_READ_8BIT_REG(op2_addr);

  /* op1 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1);
    BX_HANDLE_EXCEPTION()
    }

  sum = op1 + op2;

  /* now write sum back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, sum);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &sum);
    BX_HANDLE_EXCEPTION()
    }

  /* and write destination into source */
  BX_WRITE_8BIT_REG(op2_addr, op1);

  /* set eflags:
   * ADD affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of =  (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);
#endif
}

  INLINE void
bx_XADD_EvGv()
{
  bx_panic("XADD_EvGv:\n");
}

  INLINE void
bx_ADD_EbIb()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;

  Bit8u op2, op1, sum;
  Bit8u op2_b7, op1_b7, sum_b7;


  /* op1 is mod+r/m, op2 is an 8bit immediate */
  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  op2 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  /* op1 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1);
    BX_HANDLE_EXCEPTION()
    }

  sum = op1 + op2;

  /* now write sum back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, sum);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &sum);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * ADD affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of =  (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);
}

  INLINE void
bx_ADC_EbIb()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;

  Bit8u op2, op1, sum;
  Bit8u op2_b7, op1_b7, sum_b7;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  op2 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  /* op1 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1);
    BX_HANDLE_EXCEPTION()
    }

  sum = op1 + op2 + CF;

  /* now write sum back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, sum);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &sum);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * ADC affects the following flags: O,S,Z,A,P,C
   */
  op1_b7 = op1 & 0x80;
  op2_b7 = op2 & 0x80;
  sum_b7 = sum & 0x80;

  bx_cpu.eflags.of =  (op1_b7 == op2_b7) && (sum_b7 ^ op2_b7);
  bx_cpu.eflags.cf = (sum < op1) || (CF && sum==op1);
  bx_cpu.eflags.af = ((op1 ^ op2) ^ sum) & 0x10;
  bx_cpu.eflags.sf = (sum_b7 > 0);
  bx_cpu.eflags.zf = (sum == 0);
  SET_PF(sum);
}


  INLINE void
bx_SUB_EbIb()
{
  Bit8u op2_8, op1_8, diff_8;
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;
  Bit8u op2_b7, op1_b7, diff_b7;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  op2_8 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  /* op1_8 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - op2_8;

  /* now write diff back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, diff_8);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &diff_8);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * SUB affects the following flags: O,S,Z,A,P,C
   */
  op2_b7  = op2_8  & 0x80;
  op1_b7  = op1_8  & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of =  (op1_b7 ^ op2_b7) && (op1_b7 ^ diff_b7);
  bx_cpu.eflags.cf = (op1_8 < op2_8);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}

  INLINE void
bx_CMP_EbIb()
{
  Bit8u op2_8, op1_8, diff_8;
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;
  Bit8u op2_b7, op1_b7, diff_b7;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  op2_8 = bx_fetch_next_byte();
  BX_HANDLE_EXCEPTION()

  /* op1_8 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = op1_8 - op2_8;

  /* set eflags:
   * CMP affects the following flags: O,S,Z,A,P,C
   */
  op1_b7  = op1_8  & 0x80;
  op2_b7  = op2_8  & 0x80;
  diff_b7 = diff_8 & 0x80;

  bx_cpu.eflags.of =  (op1_b7 ^ op2_b7) && (diff_b7 ^ op1_b7);
  bx_cpu.eflags.cf = (op1_8 < op2_8);
  bx_cpu.eflags.af = ((op1_8 ^ op2_8) ^ diff_8) & 0x10;
  bx_cpu.eflags.sf = (diff_b7 > 0);
  bx_cpu.eflags.zf = (diff_8 == 0);
  SET_PF(diff_8);
}


  INLINE void
bx_ADD_EvIv()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, sum_b31;
    Bit32u op2_32, op1_32, sum_32;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, sum_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &sum_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;
    Bit16u op2_b15, op1_b15, sum_b15;


    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    sum_16 = op1_16 + op2_16;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, sum_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &sum_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}

  INLINE void
bx_ADC_EvIv()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, sum_b31;
    Bit32u op2_32, op1_32, sum_32;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    sum_32 = op1_32 + op2_32 + CF;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, sum_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &sum_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32) || (CF && sum_32==op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;
    Bit16u op2_b15, op1_b15, sum_b15;


    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    sum_16 = op1_16 + op2_16 + CF;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, sum_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &sum_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16) || (CF && sum_16==op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}


  INLINE void
bx_SUB_EvIv()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, diff_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &diff_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op2_b31 = op2_32 & 0x80000000;
    op1_b31 = op1_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - op2_16;

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, diff_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &diff_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op2_b15 = op2_16 & 0x8000;
    op1_b15 = op1_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_CMP_EvIv()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = bx_fetch_next_dword();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - op2_32;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (diff_b31 ^ op1_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    op2_16 = bx_fetch_next_word();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - op2_16;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (diff_b15 ^ op1_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_ADD_EvIb()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, sum_b31;
    Bit32u op2_32, op1_32, sum_32;

    op2_32 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, sum_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &sum_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;
    Bit16u op2_b15, op1_b15, sum_b15;


    op2_16 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    sum_16 = op1_16 + op2_16;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, sum_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &sum_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADD affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}

  INLINE void
bx_ADC_EvIb()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, sum_b31;
    Bit32u op2_32, op1_32, sum_32;

    op2_32 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    sum_32 = op1_32 + op2_32 + CF;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, sum_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &sum_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    sum_b31 = sum_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 == op2_b31) && (sum_b31 ^ op2_b31);
    bx_cpu.eflags.cf = (sum_32 < op1_32) || (CF && sum_32==op1_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ sum_32) & 0x10;
    bx_cpu.eflags.sf = (sum_b31 > 0);
    bx_cpu.eflags.zf = (sum_32 == 0);
    SET_PF(sum_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;
    Bit16u op2_b15, op1_b15, sum_b15;


    op2_16 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    sum_16 = op1_16 + op2_16 + CF;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, sum_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &sum_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * ADC affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    sum_b15 = sum_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 == op2_b15) && (sum_b15 ^ op2_b15);
    bx_cpu.eflags.cf = (sum_16 < op1_16) || (CF && sum_16==op1_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ sum_16) & 0x10;
    bx_cpu.eflags.sf = (sum_b15 > 0);
    bx_cpu.eflags.zf = (sum_16 == 0);
    SET_PF(sum_16);
    }
}

  INLINE void
bx_SUB_EvIb()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, diff_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &diff_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op2_b31 = op2_32 & 0x80000000;
    op1_b31 = op1_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (op1_b31 ^ diff_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    op2_16 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - op2_16;

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, diff_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &diff_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * SUB affects the following flags: O,S,Z,A,P,C
     */
    op2_b15 = op2_16 & 0x8000;
    op1_b15 = op1_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (op1_b15 ^ diff_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}

  INLINE void
bx_CMP_EvIb()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_b31, op1_b31, diff_b31;
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = op1_32 - op2_32;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b31 = op1_32 & 0x80000000;
    op2_b31 = op2_32 & 0x80000000;
    diff_b31 = diff_32 & 0x80000000;

    bx_cpu.eflags.of = (op1_b31 ^ op2_b31) && (diff_b31 ^ op1_b31);
    bx_cpu.eflags.cf = (op1_32 < op2_32);
    bx_cpu.eflags.af = ((op1_32 ^ op2_32) ^ diff_32) & 0x10;
    bx_cpu.eflags.sf = (diff_b31 > 0);
    bx_cpu.eflags.zf = (diff_32 == 0);
    SET_PF(diff_32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;
    Bit16u op2_b15, op1_b15, diff_b15;


    op2_16 = (Bit8s) bx_fetch_next_byte();
    BX_HANDLE_EXCEPTION()

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = op1_16 - op2_16;

    /* set eflags:
     * CMP affects the following flags: O,S,Z,A,P,C
     */
    op1_b15 = op1_16 & 0x8000;
    op2_b15 = op2_16 & 0x8000;
    diff_b15 = diff_16 & 0x8000;

    bx_cpu.eflags.of = (op1_b15 ^ op2_b15) && (diff_b15 ^ op1_b15);
    bx_cpu.eflags.cf = (op1_16 < op2_16);
    bx_cpu.eflags.af = ((op1_16 ^ op2_16) ^ diff_16) & 0x10;
    bx_cpu.eflags.sf = (diff_b15 > 0);
    bx_cpu.eflags.zf = (diff_16 == 0);
    SET_PF(diff_16);
    }
}


  INLINE void
bx_NEG_Eb()
{
  Bit8u op1_8, diff_8;
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op1_8 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  diff_8 = 0 - op1_8;

  /* now write diff back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, diff_8);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &diff_8);
    BX_HANDLE_EXCEPTION()
    }

  /* set eflags:
   * NEG affects the following flags: O,S,Z,A,P,C
   */

  bx_cpu.eflags.of = (op1_8 == 0x80);
  bx_cpu.eflags.sf = (diff_8 >= 0x80);
  bx_cpu.eflags.zf = (diff_8 == 0);
  bx_cpu.eflags.af = (op1_8 & 0x0f) > 0;
  SET_PF(diff_8);
  bx_cpu.eflags.cf = (op1_8 != 0);
}

  INLINE void
bx_NEG_Ev()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, diff_32;

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    diff_32 = 0 - op1_32;

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, diff_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &diff_32);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * NEG affects the following flags: O,S,Z,A,P,C
     */

    bx_cpu.eflags.of = (op1_32 == 0x80000000);
    bx_cpu.eflags.sf = (diff_32 >= 0x80000000);
    bx_cpu.eflags.zf = (diff_32 == 0);
    bx_cpu.eflags.af = (op1_32 & 0x0f) > 0;
    SET_PF(diff_32);
    bx_cpu.eflags.cf = (op1_32 != 0);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, diff_16;


    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    diff_16 = 0 - op1_16;

    /* now write diff back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, diff_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &diff_16);
      BX_HANDLE_EXCEPTION()
      }

    /* set eflags:
     * NEG affects the following flags: O,S,Z,A,P,C
     */

    bx_cpu.eflags.of = (op1_16 == 0x8000);
    bx_cpu.eflags.sf = (diff_16 >= 0x8000);
    bx_cpu.eflags.zf = (diff_16 == 0);
    bx_cpu.eflags.af = (op1_16 & 0x0f) > 0;
    SET_PF(diff_16);
    bx_cpu.eflags.cf = (op1_16 != 0);
    }
}


  INLINE void
bx_INC_Eb()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;

  Bit8u  op1;

  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op1 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1);
    BX_HANDLE_EXCEPTION()
    }


  op1++;

  /* now write sum back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, op1);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &op1);
    BX_HANDLE_EXCEPTION()
    }

  /* INC affect the following flags: O,S,Z,A,P */
  bx_cpu.eflags.of = (op1 == 0x80);
  bx_cpu.eflags.sf = (op1 >= 0x80);
  bx_cpu.eflags.zf = (op1 == 0);
  bx_cpu.eflags.af = (op1 & 0x0f) == 0;
  SET_PF(op1);
}

  INLINE void
bx_INC_Ev()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    op1_32++;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, op1_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    /* INC affect the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (op1_32 == 0x80000000);
    bx_cpu.eflags.sf = (op1_32 >= 0x80000000);
    bx_cpu.eflags.zf = (op1_32 == 0);
    bx_cpu.eflags.af = (op1_32 & 0x0f) == 0;
    SET_PF(op1_32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    op1_16++;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, op1_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    /* INC affect the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (op1_16 == 0x8000);
    bx_cpu.eflags.sf = (op1_16 >= 0x8000);
    bx_cpu.eflags.zf = (op1_16 == 0);
    bx_cpu.eflags.af = (op1_16 & 0x0f) == 0;
    SET_PF(op1_16);
    }
}

  INLINE void
bx_DEC_Eb()
{
  Bit8u op1_8;

  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;


  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

  /* op1_8 is a register or memory reference */
  if (op1_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(op1_addr);
    }
  else {
    /* pointer, segment address pair */
    bx_read_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  op1_8--;

  /* now write sum back to destination */
  if (op1_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(op1_addr, op1_8);
    }
  else {
    bx_write_virtual_byte(op1_seg_reg, op1_addr, &op1_8);
    BX_HANDLE_EXCEPTION()
    }

  /* DEC affects the following flags: O,S,Z,A,P */
  bx_cpu.eflags.of = (op1_8 == 0x7F);
  bx_cpu.eflags.sf = (op1_8 >= 0x80);
  bx_cpu.eflags.zf = (op1_8 == 0);
  bx_cpu.eflags.af = (op1_8 & 0x0f) == 0x0f;
  SET_PF(op1_8);
}

  INLINE void
bx_DEC_Ev()
{
  Bit32u op1_addr;
  unsigned op1_type, unused;
  bx_segment_reg_t *op1_seg_reg;

  bx_decode_exgx(&unused, &op1_addr, &op1_type, &op1_seg_reg);
  BX_HANDLE_EXCEPTION()

#if BX_CPU > 2
  if (bx_cpu.is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    op1_32--;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(op1_addr, op1_32);
      }
    else {
      bx_write_virtual_dword(op1_seg_reg, op1_addr, &op1_32);
      BX_HANDLE_EXCEPTION()
      }

    /* DEC affects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (op1_32 == 0x7FFFFFFF);
    bx_cpu.eflags.sf = (op1_32 >= 0x80000000);
    bx_cpu.eflags.zf = (op1_32 == 0);
    bx_cpu.eflags.af = (op1_32 & 0x0f) == 0x0f;
    SET_PF(op1_32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (op1_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(op1_addr);
      }
    else {
      /* pointer, segment address pair */
      bx_read_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    op1_16--;

    /* now write sum back to destination */
    if (op1_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(op1_addr, op1_16);
      }
    else {
      bx_write_virtual_word(op1_seg_reg, op1_addr, &op1_16);
      BX_HANDLE_EXCEPTION()
      }

    /* DEC affects the following flags: O,S,Z,A,P */
    bx_cpu.eflags.of = (op1_16 == 0x7FFF);
    bx_cpu.eflags.sf = (op1_16 >= 0x8000);
    bx_cpu.eflags.zf = (op1_16 == 0);
    bx_cpu.eflags.af = (op1_16 & 0x0f) == 0x0f;
    SET_PF(op1_16);
    }
}

INLINE void bx_CMPXCHG_EbGb() {bx_panic("CMPXCHG_EbGb:\n");}
INLINE void bx_CMPXCHG_EvGv() {bx_panic("CMPXCHG_EvGv:\n");}
