/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996 by Kevin P. Lawton.

BOCHS is shareware for PERSONAL USE only.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/





#include <stdio.h>
#include "bochs.h"





  INLINE void
bx_SAHF()
{
  bx_cpu.eflags.sf = (AH & 0x80) >> 7;
  bx_cpu.eflags.zf = (AH & 0x40) >> 6;
  bx_cpu.eflags.af = (AH & 0x10) >> 4;
  SET_PF_DIRECT((AH & 0x04) >> 2);
  bx_cpu.eflags.cf = (AH & 0x01);
}

  INLINE void
bx_LAHF()
{
  AH = (bx_cpu.eflags.sf ? 0x80 : 0) |
       (bx_cpu.eflags.zf ? 0x40 : 0) |
       (bx_cpu.eflags.af ? 0x10 : 0) |
       (GET_PF() ? 0x04 : 0) |
       (0x02) |
       (bx_cpu.eflags.cf ? 0x01 : 0);
}

  INLINE void
bx_CLC()
{
  bx_cpu.eflags.cf = 0;
}

  INLINE void
bx_STC()
{
  bx_cpu.eflags.cf = 1;
}

  INLINE void
bx_CLI()
{

#if BX_CPU >= 2
  if ( bx_protected_mode() && (CPL > IOPL) ) {
    /* CPL > IOPL: CPL less priveleged than IOPL */
bx_printf("CLI: CPL > IOPL\n"); /* ??? */
    bx_exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }
#endif

  bx_cpu.eflags.if_ = 0; 
}

  INLINE void
bx_STI()
{

#if BX_CPU >= 2
  if ( bx_protected_mode() && (CPL > IOPL) ) {
    /* CPL > IOPL: CPL less priveleged than IOPL */
bx_printf("STI: CPL(%u) > IOPL(%u)\n", (unsigned) CPL, (unsigned) IOPL);
    bx_exception(BX_GP_EXCEPTION, 0, 0);
    return;
    }
#endif

  if (!bx_cpu.eflags.if_) {
    bx_cpu.eflags.if_ = 1;
    bx_cpu.inhibit_interrupts = 1;
    bx_async_event = 1;
    }
}

  INLINE void
bx_CLD()
{
  bx_cpu.eflags.df = 0;
}

  INLINE void
bx_STD()
{
  bx_cpu.eflags.df = 1;
}

  INLINE void
bx_CMC()
{
  bx_cpu.eflags.cf = !bx_cpu.eflags.cf;
}


  void
bx_write_flags(Bit16u flags, Boolean change_IOPL, Boolean change_IF)
{
  bx_cpu.eflags.cf = flags & 0x01;
  SET_PF_DIRECT((flags >> 2) & 0x01);
  bx_cpu.eflags.af = (flags >> 4) & 0x01;
  bx_cpu.eflags.zf = (flags >> 6) & 0x01;
  bx_cpu.eflags.sf = (flags >> 7) & 0x01;
  bx_cpu.eflags.tf = (flags >> 8) & 0x01;
  if (bx_cpu.eflags.tf) {
    bx_async_event = 1;
bx_dbg.debugger = 1;
bx_debug(&bx_cpu.cs, bx_cpu.prev_eip);
bx_panic("write_flags(): TF set: bailing\n");
    }

  if (change_IF)
    bx_cpu.eflags.if_ = (flags >> 9) & 0x01;

  bx_cpu.eflags.df = (flags >> 10) & 0x01;
  bx_cpu.eflags.of = (flags >> 11) & 0x01;

  /* real-mode doesn't change IOPL & NT flags ??? */
  if (bx_protected_mode()) {
    if (change_IOPL)
      bx_cpu.eflags.iopl = (flags >> 12) & 0x03;
    bx_cpu.eflags.nt = (flags >> 14) & 0x01;
    }
  else { /* real mode */
    /* 286: bits 12-15 always clear
     * 386: bit 15 always clear, bits 14-12 maintain last value loaded
     */
#if BX_CPU == 2
    bx_cpu.eflags.iopl = 0;
    bx_cpu.eflags.nt = 0;
#elif BX_CPU >= 3
    /* values shouldn't be used in real mode, but are loaded into eflags */
    bx_cpu.eflags.iopl = (flags >> 12) & 0x03;
    bx_cpu.eflags.nt = (flags >> 14) & 0x01;
#endif
    }
}

#if BX_CPU >= 3
  void
bx_write_eflags(Bit32u eflags, Boolean change_IOPL, Boolean change_IF)
{
if (eflags & 0x00020000)
  bx_panic("write_eflags: VM bit set\n");


  bx_cpu.eflags.cf = eflags & 0x01;
  SET_PF_DIRECT((eflags >> 2) & 0x01);
  bx_cpu.eflags.af = (eflags >> 4) & 0x01;
  bx_cpu.eflags.zf = (eflags >> 6) & 0x01;
  bx_cpu.eflags.sf = (eflags >> 7) & 0x01;
  bx_cpu.eflags.tf = (eflags >> 8) & 0x01;
  if (bx_cpu.eflags.tf) {
    bx_async_event = 1;
    bx_panic("write_eflags(): TF set: bailing\n");
    }

  if (change_IF)
    bx_cpu.eflags.if_ = (eflags >> 9) & 0x01;

  bx_cpu.eflags.df = (eflags >> 10) & 0x01;
  bx_cpu.eflags.of = (eflags >> 11) & 0x01;

  /* real-mode doesn't change IOPL & NT flags ??? */
  if (bx_protected_mode()) {
    if (change_IOPL)
      bx_cpu.eflags.iopl = (eflags >> 12) & 0x03;
    bx_cpu.eflags.nt = (eflags >> 14) & 0x01;
    }
  else { /* real mode */
    /* 386+: bit 15 always clear, bits 14-12 maintain last value loaded
     */
    /* values shouldn't be used in real mode, but are loaded into eflags */
    bx_cpu.eflags.iopl = (eflags >> 12) & 0x03;
    bx_cpu.eflags.nt = (eflags >> 14) & 0x01;
    }

  /* RF & VM flags not affected by this function.  They should be
   * manipulated directly by other instructions ??? */
  /* ??? what to do with ID,VIP,VIF,AC,VM,RF */
  /* popf/popfd does not affect vm & rf (bits 16&17) */
}
#endif /* BX_CPU >= 3 */

  Bit16u
bx_read_flags(void)
{
  Bit16u flags;

#if BX_WEIRDISMS
  if (bx_cpu.eflags.cf>1 || 
      bx_cpu.eflags.zf>1 || 
      bx_cpu.eflags.sf>1 || 
      bx_cpu.eflags.tf>1 || 
      bx_cpu.eflags.if_>1 || 
      bx_cpu.eflags.df>1 || 
      bx_cpu.eflags.of>1 || 
      bx_cpu.eflags.iopl>3 || 
      bx_cpu.eflags.nt>1) {
    bx_printf("cf = %u\n", (unsigned) bx_cpu.eflags.cf);
    bx_printf("af = %u\n", (unsigned) bx_cpu.eflags.af);
    bx_printf("zf = %u\n", (unsigned) bx_cpu.eflags.zf);
    bx_printf("sf = %u\n", (unsigned) bx_cpu.eflags.sf);
    bx_printf("tf = %u\n", (unsigned) bx_cpu.eflags.tf);
    bx_printf("if_ = %u\n", (unsigned) bx_cpu.eflags.if_);
    bx_printf("df = %u\n", (unsigned) bx_cpu.eflags.df);
    bx_printf("of = %u\n", (unsigned) bx_cpu.eflags.of);
    bx_printf("iopl = %u\n", (unsigned) bx_cpu.eflags.iopl);
    bx_printf("nt = %u\n", (unsigned) bx_cpu.eflags.nt);
    bx_panic("read_flags(): bad flag values\n");
    }
#endif

  flags = (bx_cpu.eflags.cf) |
          (bx_cpu.eflags.bit1 << 1) |
          ((GET_PF()) << 2) |
          (bx_cpu.eflags.bit3 << 3) |
          ((bx_cpu.eflags.af>0) << 4) |
          (bx_cpu.eflags.bit5 << 5) |
          ((bx_cpu.eflags.zf>0) << 6) |
          ((bx_cpu.eflags.sf>0) << 7) |
          (bx_cpu.eflags.tf << 8) |
          (bx_cpu.eflags.if_ << 9) |
          (bx_cpu.eflags.df << 10) |
          ((bx_cpu.eflags.of>0) << 11) |
          (bx_cpu.eflags.iopl << 12) |
          (bx_cpu.eflags.nt << 14) |
          (bx_cpu.eflags.bit15 << 15);

  /* 8086: bits 12-15 always set to 1.
   * 286: in real mode, bits 12-15 always cleared.
   * 386+: real-mode: bit15 cleared, bits 14..12 are last loaded value
   *       protected-mode: bit 15 clear, bit 14 = last loaded, IOPL?
   */
#if BX_CPU < 2
  flags |= 0xF000;  /* 8086 nature */
#elif BX_CPU == 2
  if (bx_real_mode()) {
    flags &= 0x0FFF;  /* 80286 in real mode nature */
    }
#else /* 386+ */
  if (bx_real_mode()) {
    flags &= 0x7FFF;  /* 80386 in real mode nature */
    }
#endif

  return(flags);
}


#if BX_CPU >= 3
  Bit32u
bx_read_eflags(void)
{
  Bit32u eflags;

#if BX_CPU >= 4
  bx_panic("read_eflags(): not implemented for 80486+\n");
#endif

#if BX_WEIRDISMS
  if (bx_cpu.eflags.cf>1 || 
      bx_cpu.eflags.zf>1 || 
      bx_cpu.eflags.sf>1 || 
      bx_cpu.eflags.tf>1 || 
      bx_cpu.eflags.if_>1 || 
      bx_cpu.eflags.df>1 || 
      bx_cpu.eflags.of>1 || 
      bx_cpu.eflags.iopl>3 || 
      bx_cpu.eflags.nt>1) {
    bx_printf("cf = %u\n", (unsigned) bx_cpu.eflags.cf);
    bx_printf("af = %u\n", (unsigned) bx_cpu.eflags.af);
    bx_printf("zf = %u\n", (unsigned) bx_cpu.eflags.zf);
    bx_printf("sf = %u\n", (unsigned) bx_cpu.eflags.sf);
    bx_printf("tf = %u\n", (unsigned) bx_cpu.eflags.tf);
    bx_printf("if_ = %u\n", (unsigned) bx_cpu.eflags.if_);
    bx_printf("df = %u\n", (unsigned) bx_cpu.eflags.df);
    bx_printf("of = %u\n", (unsigned) bx_cpu.eflags.of);
    bx_printf("iopl = %u\n", (unsigned) bx_cpu.eflags.iopl);
    bx_printf("nt = %u\n", (unsigned) bx_cpu.eflags.nt);
    bx_panic("read_flags(): bad flag values\n");
    }
#endif

  eflags = (bx_cpu.eflags.cf) |
          (bx_cpu.eflags.bit1 << 1) |
          ((GET_PF()) << 2) |
          (bx_cpu.eflags.bit3 << 3) |
          ((bx_cpu.eflags.af>0) << 4) |
          (bx_cpu.eflags.bit5 << 5) |
          ((bx_cpu.eflags.zf>0) << 6) |
          ((bx_cpu.eflags.sf>0) << 7) |
          (bx_cpu.eflags.tf << 8) |
          (bx_cpu.eflags.if_ << 9) |
          (bx_cpu.eflags.df << 10) |
          ((bx_cpu.eflags.of>0) << 11) |
          (bx_cpu.eflags.iopl << 12) |
          (bx_cpu.eflags.nt << 14) |
          (bx_cpu.eflags.bit15 << 15) |
          (bx_cpu.eflags.rf << 16) |
          (bx_cpu.eflags.vm << 17);

  /*
   * 386+: real-mode: bit15 cleared, bits 14..12 are last loaded value
   *       protected-mode: bit 15 clear, bit 14 = last loaded, IOPL?
   */
  if (bx_real_mode()) {
    eflags &= 0x7FFF;  /* 80386 in real mode nature */
    }

  return(eflags);
}
#endif /* BX_CPU >= 3 */


  INLINE void
bx_PUSHF_Fv()
{
#if BX_CPU >= 3
  if (bx_virtual_mode() == 0) { /* not in virtual-8086 mode */
    if (bx_cpu.is_32bit_opsize) {
      bx_push_32(bx_read_eflags() & 0x00fcffff);
      /* will return after error anyway */
      }
    else {
#endif
      bx_push_16(bx_read_flags());
      /* will return after error anyway */
#if BX_CPU >= 3
      }
    }
  else { /* v8086 mode */
    bx_panic("pushf(): not implemented for v8086 mode\n");
    }
#endif
}


  INLINE void
bx_POPF_Fv()
{
#if BX_CPU >= 3
  if (bx_virtual_mode()) {
    bx_panic("popf_fv(): not supported in virtual mode\n");
    return;
    }

  if (bx_cpu.is_32bit_opsize) {
    Bit32u eflags;

    bx_pop_32(&eflags);
    BX_HANDLE_EXCEPTION()

    eflags &= 0x00277fd7;
    if (bx_protected_mode()) {
      bx_write_eflags(eflags, CPL==0, CPL<=IOPL);
      }
    else { /* real mode */
      bx_write_eflags(eflags, 1, 1);
      }
    }
  else
#endif /* BX_CPU >= 3 */
    { /* 16 bit opsize */
    Bit16u flags;

    bx_pop_16(&flags);
    BX_HANDLE_EXCEPTION()

    if (bx_protected_mode()) {
      bx_write_flags(flags, CPL==0, CPL<=IOPL);
      }
    else { /* real mode */
      bx_write_flags(flags, 1, 1);
      }
    }
}
