/*
Copyright Notice
================
BOCHS is Copyright 1994,1995,1996 by Kevin P. Lawton.

BOCHS is shareware for PERSONAL USE only.

For more information, read the file 'LICENSE' included in the bochs
distribution.  If you don't have access to this file, or have questions
regarding the licensing policy, the author may be contacted via:

    US Mail:  Kevin Lawton
              528 Lexington St.
              Waltham, MA 02154

    EMail:    bochs@world.std.com
*/



#define BX_MEMORY_C 1



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ??? */
#include <ctype.h>

#include "bochs.h"
#include "iodev/iodev.h"



Bit8u bx_phy_memory[BX_PHY_MEM_SIZE];





/* Address line 20 control:
   1 = enabled: extended memory is accessible
   0 = disabled: A20 address line is forced low to simulate an 8088 address map
 */
Boolean bx_enable_a20 = 1;

/* start out masking physical memory addresses to:
 *   8086:      20 bits
 *    286:      24 bits
 *    386:      32 bits
 * when A20 line is disabled, mask physical memory addresses to:
 *    286:      20 bits
 *    386:      20 bits
 */

#if BX_CPU < 2
  Bit32u  bx_a20_mask   =    0xfffff;
#elif BX_CPU == 2
  Bit32u  bx_a20_mask   =   0xffffff;
#else /* 386+ */
  Bit32u  bx_a20_mask   = 0xffffffff;
#endif


#define A20ADDR(x) ( (x) & bx_a20_mask )


extern Bit8u bx_io_handler_id[0x10000];  /* 64K */
extern bx_io_handler_t bx_io_handler[BX_MAX_IO_DEVICES];


  char *
bx_strseg(bx_segment_reg_t *seg)
{
  if (seg == &bx_cpu.es) return("ES");
  else if (seg == &bx_cpu.cs) return("CS");
  else if (seg == &bx_cpu.ss) return("SS");
  else if (seg == &bx_cpu.ds) return("DS");
  else if (seg == &bx_cpu.fs) return("FS");
  else if (seg == &bx_cpu.gs) return("GS");
  else {
    bx_printf("undefined segment passed to bx_strseg()!\n");
    return("??");
    }
}

  void
bx_init_memory(int memsize)
{
  memset(bx_phy_memory, 0x00, memsize);
}

  void
bx_virtual_block_write(bx_segment_reg_t *seg, Bit16u offset, Bit32u length,
  void *data)
{
  int i;
  Bit8u temp8;

#if BX_CPU >= 2
  if (bx_protected_mode()) {
    bx_panic(
      "virtual_block_write() not yet supported for protected mode\n");
    }
#endif

  for (i=0; i < length; i++) {
    temp8 = *(((Bit8u*) data) + i);
    bx_write_virtual_byte(seg, offset + i, &temp8);
    if (bx_cpu.errno) {
      bx_panic("virtual_block_write(): errno set\n");
      }
    }
}

  void
bx_virtual_block_read(bx_segment_reg_t *seg, Bit16u offset, Bit32u length,
  void *data)
{
  int i;
  Bit8u temp8;

#if BX_CPU >= 2
  if (bx_protected_mode()) {
    bx_panic(
      "virtual_block_read() not yet supported for protected mode\n");
    }
#endif

  for (i=0; i < length; i++) {
    bx_read_virtual_byte(seg, offset + i, &temp8);
    *( ((Bit8u*) data) + i ) = temp8;
    if (bx_cpu.errno) {
      bx_panic("virtual_block_read(): errno set\n");
      }
    }
}



#if BX_CPU >= 2
  void
bx_access_virtual(bx_segment_reg_t *seg, Bit32u offset, Bit8u length, 
    Bit8u rw, void *data)
{
  Bit32u upper_limit;

  if ( bx_protected_mode() ) {
    if ( seg->cache.valid==0 ) {
      bx_printf("seg = %s\n", bx_strseg(seg));
      bx_printf("seg->selector.value = %04x\n", (unsigned) seg->selector.value);
      bx_printf("access_virtual: valid bit = 0\n");
bx_printf("CS: %04x\n", (unsigned) bx_cpu.cs.selector.value);
bx_printf("IP: %04x\n", (unsigned) bx_cpu.prev_eip);
      bx_exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }

    if (seg->cache.p == 0) { /* not present */
bx_printf("access_virtual(): segment not present\n");
      bx_exception(bx_int_number(seg), 0, 0);
      return;
      }

    switch ( seg->cache.type ) {
      case 0: case 1: /* read only */
      case 10: case 11: /* execute/read */
      case 14: case 15: /* execute/read-only, conforming */
        if (rw==BX_WRITE) {
bx_printf("access_virtual(): write to read-only\n");
          bx_exception(bx_int_number(seg), 0, 0);
          return;
          }
        if ( (offset+length-1) > seg->cache.u.segment.limit_scaled ) {
bx_printf("access_virtual(): write beyond limit\n");
          bx_exception(bx_int_number(seg), 0, 0);
          return;
          }
        break;

      case 2: case 3: /* read/write */
        if ( (offset+length-1) > seg->cache.u.segment.limit_scaled ) {
bx_printf("access_virtual(): write beyond limit\n");
          bx_exception(bx_int_number(seg), 0, 0);
          return;
          }
        break;

      case 4: case 5: /* read only, expand down */
        if (rw==BX_WRITE) {
bx_printf("access_virtual(): read-only\n");
          bx_exception(bx_int_number(seg), 0, 0);
          return;
          }
        if (seg->cache.u.segment.d_b)
          upper_limit = 0xffffffff;
        else
          upper_limit = 0x0000ffff;
        if ( (offset <= seg->cache.u.segment.limit_scaled) ||
             (offset > upper_limit) ||
             ((upper_limit - offset) < (length - 1)) ) {
bx_printf("access_virtual(): write beyond limit\n");
          bx_exception(bx_int_number(seg), 0, 0);
          return;
          }
        break;

      case 6: case 7: /* read write, expand down */
        if (seg->cache.u.segment.d_b)
          upper_limit = 0xffffffff;
        else
          upper_limit = 0x0000ffff;
        if ( (offset <= seg->cache.u.segment.limit_scaled) ||
             (offset > upper_limit) ||
             ((upper_limit - offset) < (length - 1)) ) {
bx_printf("access_virtual(): write beyond limit\n");
          bx_exception(bx_int_number(seg), 0, 0);
          return;
          }
        break;

      case 8: case 9: /* execute only */
      case 12: case 13: /* execute only, conforming */
        /* can't read or write an execute-only segment */
bx_printf("access_virtual(): execute only\n");
        bx_exception(bx_int_number(seg), 0, 0);
        return;
        break;
      }

    /* all checks OK */
#if BX_CPU >= 3
    /* need to look at access crossing multiple page boundary ??? */
    if (bx_cpu.cr0.pg)
      bx_access_linear(seg->cache.u.segment.base + offset,
        length, CPL, rw, (void *) data);
    else
#endif
      bx_access_physical(seg->cache.u.segment.base + offset, length, rw, data);
    /* will return if error anyway */
    return;
    }

  else { /* real mode */
    if ( (offset + length - 1)  >  seg->cache.u.segment.limit_scaled) {
      bx_printf("access_virtual() SEG EXCEPTION:  %x:%x + %x\n",
        (unsigned) seg->selector.value, (unsigned) offset, (unsigned) length);
      if (seg == &bx_cpu.ss) bx_exception(BX_SS_EXCEPTION, 0, 0);
      else bx_exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
#if BX_CPU >= 3
    if (bx_cpu.cr0.pg)
      bx_access_linear(seg->cache.u.segment.base + offset,
        length, CPL, rw, (void *) data);
    else
#endif
      bx_access_physical(seg->cache.u.segment.base + offset, length, rw, data);
    /* will return if error anyway */
    return;
    }
}
#endif /* BX_CPU >= 2 */




  void
bx_access_physical(Bit32u addr, Bit8u len, Bit8u rw, void *data)
{
  Bit32u i;
  Bit8u  *data_ptr;

  addr = A20ADDR(addr);

  if ( (addr + len) > BX_PHY_MEM_SIZE ) {
    if (rw==BX_READ)
      * (Bit8u *) data = 0xff;
    return;
#if 0
    bx_panic("access beyond physical memory boundary! (%08x+%02x)\n",
      (unsigned) addr, (unsigned) len);
#endif
    }



#ifdef LITTLE_ENDIAN
  data_ptr = (Bit8u *) data;

  for (i=1; i<=len; i++, addr++, data_ptr++) {
#else /* BIG_ENDIAN */
  data_ptr = (Bit8u *) data + sizeof(Bit8u) * (len - 1);

  for (i=1; i<=len; i++, addr++, data_ptr--) {
#endif

    if (rw == BX_READ)
      *(Bit8u *) data_ptr = bx_phy_memory[addr];
    else {
      bx_phy_memory[addr] = *(Bit8u *) data_ptr;
      }

    } /* for (i=... */
}


  void
bx_set_interrupt_vector(int interrupt, Bit32u address)
{
  Bit16u ip, cs;


  ip = (Bit16u) (address & 0xFFFF);
  cs = (Bit16u) ((address & 0xF0000) >> 4);
  bx_access_physical(interrupt*4, 2, BX_WRITE, &ip);
  bx_access_physical(interrupt*4 + 2, 2, BX_WRITE, &cs);
}


/* bx_inp()
 *
 * Read a byte of data from the IO memory address space
 */

  Bit32u
bx_inp(Bit16u addr, unsigned io_len)
{
  Bit8u handle;
  Bit32u ret;

  handle = bx_io_handler_id[addr];
  ret = (* bx_io_handler[handle].read_funct)((Bit32u) addr, io_len);

  if (bx_dbg.io)
    bx_printf("IO: inp(%04x) = %02x\n", (unsigned) addr, (unsigned) ret);
  return(ret);
}


/* bx_outp()
 *
 * Write a byte of data to the IO memory address space.
 */

  void
bx_outp(Bit16u addr, Bit32u value, unsigned io_len)
{
  Bit8u handle;

  handle = bx_io_handler_id[addr];
  (* bx_io_handler[handle].write_funct)((Bit32u) addr, value, io_len);

  if (bx_dbg.io)
    bx_printf("IO: outp(%04x) = %02x\n", (unsigned) addr, (unsigned) value);
}


  void
bx_set_enable_a20(Bit8u value)
{
#if BX_CPU < 2
    bx_panic("set_enable_a20() called: 8086 emulation\n");
#else

  if (value) {
    bx_enable_a20 = 1;
#if BX_CPU == 2
    bx_a20_mask   = 0xffffff;   /* 286: enable all 24 address lines */
#else /* 386+ */
    bx_a20_mask   = 0xffffffff; /* 386: enable all 32 address lines */
#endif
    }
  else {
    bx_enable_a20 = 0;
    bx_a20_mask   = 0x0fffff;   /* mask off all but lower 20 address lines */
    /* ??? maybe should only mask off A20 line */
    }

  if (bx_dbg.a20)
    bx_printf("A20: set() = %u\n", (unsigned) bx_enable_a20);
#endif
}

  Boolean
bx_get_enable_a20(void)
{
  if (bx_dbg.a20)
    bx_printf("A20: get() = %u\n", (unsigned) bx_enable_a20);

  if (bx_enable_a20) return(1);
  else return(0);
}
